
/* 
 * SuperCollider3 source file "SFTime.sc" 
 * Written by Jens Gulden, jgulden@cs.tu-berlin.de.
 * Licensed under the GNU General Public License (GPL),
 * this software comes with NO WARRANTY.
 */

// --- class SFTime -----------------------------------------------------------
// 
SFTime {

    // --- attributes

    classvar <>mode_REL = 0; // type  int
    classvar <>mode_BEATS = 1; // type  int
    classvar <>mode_SECS = 2; // type  int
    classvar <>mode_FRAMES = 3; // type  int
    classvar <>formatDecimals = 3; // type  int
    var <>value; // type  Float
    var <>mode; // type  int


    // --- relationships

    var <>relAnker; // 0..1-relation to type  SFAnker



    // --- new(rel) : SFTime --------------------------------------------------
    //      
    *new { arg rel; // type SFAnker         
        // the relative anker also determines inside of which slice this time occurs
        ^super.new.relAnker_(rel); // type  SFTime
    } // end new        


    // --- rel(rel) : SFTime --------------------------------------------------
    //      
    rel { arg rel; // type Float         
        value = rel;
        mode = mode_REL;
    } // end rel        


    // --- beats(beats) : SFTime ----------------------------------------------
    //      
    beats { arg beats; // type Float         
        value = beats;
        mode = mode_BEATS;
    } // end beats        


    // --- secs(secs) : SFTime ------------------------------------------------
    //      
    secs { arg secs; // type Float         
        value = secs;
        mode = mode_SECS;
    } // end secs        


    // --- frames(frames) : SFTime --------------------------------------------
    //      
    frames { arg frames; // type int         
        value = frames;
        mode = mode_FRAMES;
    } // end frames        


    // --- parse(time) :  -----------------------------------------------------
    //
    // <p>Parses an object describing as an instance of SFTime.</p>
    // <p>Parameter time may be of three alternative types:</p>
    // <ul type="disc">
    //     <li>time.isKindOf(Double): the numeric value is treated as a relative length value and a new instance of SFTime gets created with it and returned</li>
    //     <li>time.isKindOf(String): the string gets parsed and a new instance of SFTime gets created and returned</li>
    //     <li>time.isKindOf(SFTime): time is returned</li>
    // </ul>
    // <p></p>
    // 
    // 
    // 
    // 
    // @param time             
    parse { arg time; // type Object         
        var factor, string;
        factor = 1.0;
        //"*** PARSE: ".post; time.postln;
        case 
        { time.isKindOf(Number) } {
            mode = mode_REL;
            value = time;            
        }
        { time.isKindOf(SFTime) } {
            this.set(time);
        }
        { time.isKindOf(String) } {
            //"RRRRRRRRRRRRRRRRRR".postln;
            string = Toolbox.trim(time);
            mode = mode_REL; // assume sufffix-less float-value
            case 
                { string.endsWith("%") } {
                    string = string.copyFromStart(string.size-1);
                    mode = mode_REL;
                    factor = 0.01;
                }
                { string.endsWith("ms") } {
                    string = string.copyFromStart(string.size-2);
                    mode = mode_SECS;
                    factor = 0.001;
                }
                { string.endsWith("s") } {
                    string = string.copyFromStart(string.size-1);
                    mode = mode_SECS;
                }
                { string.endsWith("~") } {
                    string = string.copyFromStart(string.size-1);
                    mode = mode_BEATS;
                }
                { string.endsWith("#") } {
                    string = string.copyFromStart(string.size-1);
                    mode = mode_FRAMES;
                };
            value = string.asFloat * factor;
        }
        { true } { // else
            time.post; " - error: cannot parse time".die;
        }
        ^this;
    } // end parse        


    // --- as(modeNum) : SFTime -----------------------------------------------
    //      
    as { arg modeNum; // type int         
        var a, b;
        
        // PREMATURE, will not always work
        
        if ( mode == modeNum , {
             ^this;
         },{
             switch ( modeNum ,
                 mode_REL, {
                     if ( this.mode == mode_SECS , {
                         ^SFTime.new(this.relAnker).mode_(mode_REL).value_(this.value / this.relAnker.ownerSlice.length.asSecs.value);
                     },{
                         ^SFTime.new(this.relAnker).mode_(mode_REL).value_(this.asFrames.value / this.relAnker.ownerSlice.length.asFrames.value);
                     });
                 },
                 mode_BEATS, {
                     a = this.asRel;
                     //"!!a = ".post; a.value.postln;
                     a = a.absolute.value;
                     //"!a = ".post; a.postln;
                     b = this.relAnker.ownerSlice.masterslice.master.beats;
                     //"!b = ".post; b.postln;
                     ^SFTime.new(this.relAnker).mode_(mode_BEATS).value_(a * b);
                 },
                 mode_SECS, {
                     ^SFTime.new(this.relAnker).mode_(mode_SECS).value_(this.asFrames.value / this.relAnker.ownerSlice.masterslice.master.sampleRate );
                 },
                 mode_FRAMES, {
                     ^SFTime.new(this.relAnker).mode_(mode_FRAMES).value_(this.asRel.absolute.value * this.relAnker.ownerSlice.masterslice.master.numFrames);
                 }
             );
         });
    } // end as        


    // --- asRel() : SFTime ---------------------------------------------------
    //     
    asRel {        
        ^this.as(SFTime.mode_REL); // type  SFTime
    } // end asRel        


    // --- asBeats() : SFTime -------------------------------------------------
    //     
    asBeats {        
        ^this.as(SFTime.mode_BEATS); // type  SFTime
    } // end asBeats        


    // --- asSecs() : SFTime --------------------------------------------------
    //     
    asSecs {        
        ^this.as(SFTime.mode_SECS); // type  SFTime
    } // end asSecs        


    // --- asFrames() : SFTime ------------------------------------------------
    //     
    asFrames {        
        ^this.as(SFTime.mode_FRAMES); // type  SFTime
    } // end asFrames        


    // --- absolute() : SFTime ------------------------------------------------
    //     
    absolute {        
        var a;
        //"$$$$ ABSOLUTE: ".post; relAnker.identityHash.postln;
        if ( relAnker != nil, {
            //"%%%%AAA".post; relAnker.postln;
            a = relAnker.ownerSlice.masterslice.start;
            //"%%%% slice: ".post; a.postln;
            //a = a.masterslice;
            //"%%%%aa ".post; a.postln;
            //a = a.start;
            //"%%%%aaa".postln;
            ^( this.relativeTo( a ) ); // type  SFTime
        },{
            //"%%%%BBB".postln;
            ^this;
        });
    } // end absolute        


    // --- relativeTo(anker) : SFTime -----------------------------------------
    //      
    relativeTo { arg anker; // type SFAnker         
        // the result SFTime has
        // - the same _mode_ as this SFTime
        // - _relAnker_ set to anker, and at the same time modified _value_
        //   to represent the same position in the waveform (at least until
        //   tempo and samplerate remain unchanged), but ankered differently
        var strt, a, t, s; 
        //"REL-ANKER-SLICE: ".post; relAnker.ownerSlice.postln;
        strt = relAnker.ownerSlice.start;
        //"RELATIVE".postln;
        if ( anker === relAnker , {
                //"&&& AAA &&&".postln;
                ^this;
        },{
            if ( anker.slice == relAnker.slice , { // relative to an anker in the same slice
                //"&&& BBB &&&".postln;
                // determine via difference
                ^this.diff(anker.time); // will create a new instance, type SFTime
            },{
                //"&&& CCC &&&".postln;
                // determine in two steps
                t = relAnker.time.as(this.mode).relativeTo(relAnker.slice.start); // 1. inside this slice: normalize the anker to which this SFTime is relative
                s = relAnker.slice.start.time.relativeTo(anker); // 2. get start anker of this slice relative to the requested anker
                t = t.sum(this); // t is now SFTime relative to the start anker of this time's slice (the usual case, in most cases relAnker had already been the start anker, thus t had been 0)
                // now create sum of s and t, based on absolute values
                a = s.asFrames.sum(t.asFrames).as(this.mode);
                ^a; // type  SFTime
            });
        });
    } // end relativeTo        


    // --- plus(t) : SFTime ---------------------------------------------------
    //      
    plus { arg t; // type SFTime         
        value = value + t.as(mode).value;
        ^this; // type  SFTime
    } // end plus        


    // --- minus(t) : SFTime --------------------------------------------------
    //      
    minus { arg t; // type SFTime         
        value = value - t.as(mode).value;
        ^this; // type  SFTime
    } // end minus        


    // --- sum(t) : SFTime ----------------------------------------------------
    //      
    sum { arg t; // type SFTime         
        ^this.clone.plus(t); // type  SFtime
    } // end sum        


    // --- diff(t) : SFTime ---------------------------------------------------
    //      
    diff { arg t; // type SFTime         
        ^this.clone.minus(t); // type  SFtime
    } // end diff        


    // --- clone() : SFTime ---------------------------------------------------
    //     
    clone {        
        var t;
        t = SFTime.new(this.relAnker);
        t.value = this.value;
        t.mode = this.mode;
        ^t; // type  SFTime
    } // end clone        


    // --- post() : void ------------------------------------------------------
    //   
    post {        
        var v, s;
        v = this.value;
        if ( (this.mode == mode_REL), {
            v = v * 100;
        });
        s = Toolbox.numberFormat(v, SFTime.formatDecimals, SFTime.formatDecimals, 3);
        (s ++ switch (this.mode, 
             mode_BEATS, {"~"},
             mode_SECS, {"s"},
             mode_FRAMES, {"#"},
             mode_REL, {"%"})).post; 
    } // end post        


    // --- postln() : void ----------------------------------------------------
    //   
    postln {        
        this.post; "".postln;
    } // end postln        


    // --- set(t) : void ------------------------------------------------------
    //     
    set { arg t; // type SFTime         
        this.value = t.as(this.mode).value;
    } // end set        


} // end SFTime
