
/* 
 * SuperCollider3 source file "Toolbox.sc" 
 * Written by Jens Gulden, jgulden@cs.tu-berlin.de.
 * Licensed under the GNU General Public License (GPL),
 * this software comes with NO WARRANTY.
 */

// --- class Toolbox ----------------------------------------------------------
// 
Toolbox {

    // --- trimLeft(s) : String -----------------------------------------------
    //      
    *trimLeft { arg s; // type String         
        var i = 0;
        
        if ( s.size > 0 , {
            if ( s[0].isSpace , {
                while ( { (i < (s.size-1)) && (s[i].isSpace) } , { i = i + 1 } );
                if ( ((i < (s.size-1)) || (not (s[i].isSpace))) , {
                    ^s.copyToEnd(i);
                },{
                    ^"";
                });
            },{
                ^s;
            });
        },{
            ^"";
        });
    } // end trimLeft        


    // --- trimRight(s) : String ----------------------------------------------
    //      
    *trimRight { arg s; // type String         
        var i;
        i = s.size - 1;
        if ( i >= 0 , {
            if ( s[i].isSpace , {
                while ( { ( i > 0 ) && (s[i].isSpace) } , { i = i - 1 } );
                if ( ((i > 0) || (not (s[i].isSpace))) , {
                    ^s.copyFromStart(i);
                },{
                    ^"";
                });
            },{
                ^s;
            });
        },{
            ^"";
        });
    } // end trimRight        


    // --- trim(s) : String ---------------------------------------------------
    //      
    *trim { arg s; // type String         
        ^Toolbox.trimLeft (Toolbox.trimRight( s ) );
    } // end trim        


    // --- repeat(s, n) : String ----------------------------------------------
    //       
    *repeat { arg s, n; // types String, int          
        if ( n <= 0 , {
            ^"";
        },{
            ^(s ++ Toolbox.repeat( s, n - 1 )); // recursion
        });
    } // end repeat        


    // --- isMultiline(s) : boolean -------------------------------------------
    //      
    *isMultiline { arg s; // type String         
        ^( s.find(Char.nl) != nil ); // type  boolean
    } // end isMultiline        


    // --- split(s, delim) : List ---------------------------------------------
    //       
    *split { arg s, delim; // types String, String          
        var l;
        var f;
        var pos;
        
        l = List.new;
        f = { arg s;
            pos = s.find(delim);
            if ( pos != nil , {
                l.add(s.copyFromStart(pos-1));
                f.value(s.copyToEnd(pos+delim.size)); // recursion
            },{
                l.add(s);
            });
        };        
        f.value(s);
        ^l; // type List
    } // end split        


    // --- splitLines(s) : List -----------------------------------------------
    //      
    *splitLines { arg s; // type String         
        ^Toolbox.split(s, Char.nl);
    } // end splitLines        


    // --- join(l, delim) : String --------------------------------------------
    //       
    *join { arg l, delim; // types List, String          
        var s = nil;
        
        l.do({ arg e;
            if ( s == nil , {
                s = e;
            },{
                s = s ++ delim ++ e;
            });
        });
        ^s;
    } // end join        


    // --- joinLines(l) : String ----------------------------------------------
    //      
    *joinLines { arg l; // type List         
        ^Toolbox.join(l, Char.nl); // type  String
    } // end joinLines        


    // --- indent(s, shift) : String ------------------------------------------
    //       
    *indent { arg s, shift; // types String, int          
        var lines;
        var ind;
        
        lines = Toolbox.splitLines(s);
        ind = Toolbox.repeat(" ", shift);
        lines.do({ arg l, index;
            lines.put(index, (ind ++ Toolbox.trimLeft(l)));
        });
        ^Toolbox.joinLines(lines);
    } // end indent        


    // --- unindent(s, shift) : String ----------------------------------------
    //       
    *unindent { arg s, shift; // types String, int          
        // if shift==-1: unindent to maximum possible (at least one line will start without blanks, oters are shifted relatively)
        var lines;
        var spaces;
        
        lines = Toolbox.splitLines(s);
        if ( shift == -1 , { // first find shift value: minimum number of blanks at the beginning of lines
            lines.do({ arg line;
                spaces = line.size - Toolbox.trimLeft(line).size;
                if ( ( (shift == -1) || (spaces < shift) ) , {
                    shift = spaces;
                });
            });
        });
        
        if ( shift > 0 , {
            lines.do({ arg line, index;
                line = line.copyToEnd(shift); // cut left part
                lines.put(index, line);
            });
        });
        
        ^Toolbox.joinLines(lines); // type  String
    } // end unindent        


    // --- indentAbs(s, shift) : String ---------------------------------------
    //       
    *indentAbs { arg s, shift; // types String, int          
        // indent to absolute shifting, i.e. independently of original indentation        
        ^Toolbox.indent( Toolbox.unindent(s, -1), shift ); // type  String
    } // end indentAbs        


    // --- hasInstanceVar(o, varname) : boolean -------------------------------
    //       
    *hasInstanceVar { arg o, varname; // types Object, String          
        var n;
        
        n = o.class.instVarNames;
        if ( n != nil , {
            ^n.includes(varname.asSymbol); // type  boolean
        },{
            ^false;
        });
    } // end hasInstanceVar        


    // --- hasClassVar(o, varname) : boolean ----------------------------------
    //       
    *hasClassVar { arg o, varname; // types Object, String          
        var n;
        
        n = o.class.classVarNames;
        if ( n != nil , {
            ^n.includes(varname.asSymbol); // type  boolean
        },{
            ^false;
        });
    } // end hasClassVar        


    // --- hasVar(o, varname) : boolean ---------------------------------------
    //       
    *hasVar { arg o, varname; // types Object, String          
        //"++++ hasVar: ".post; varname.post; o.class.instVarNames.postln;
        ^((Toolbox.hasInstanceVar(o, varname)) || (Toolbox.hasClassVar(o, varname))); // type  boolean
    } // end hasVar        


    // --- getVar(o, varname) : Object ----------------------------------------
    //       
    *getVar { arg o, varname; // types Object, String          
        var f;
        //"GET VAR ".post; o.post; ".".post; varname.postln;
        f = compile("{arg a; a."++varname++"}");
        ^(f.value.value(o));
    } // end getVar        


    // --- setVar(o, varname, value) : void -----------------------------------
    //       
    *setVar { arg o, varname, value; // types Object, String, Object           
        var f;
        //"SET VAR ".post; o.post; ".".post; varname.post; "=".post; value.postln;
        f = compile("{ arg a, v; a."++varname++"=v }");
        (f.value.value(o));
    } // end setVar        


    // --- numberFormat(num, maxDec, minDec, minWhole, leadFill) : String -----
    //          
    *numberFormat { arg num, maxDec, minDec, minWhole, leadFill; // types SimpleNumber, int, int, int, String             
        var v, dec, s, inspos;
        
        v = num;
        if ( (maxDec == nil), {
            maxDec = 2;
        });
        if ( (minDec == nil), {
            minDec = maxDec;
        });
        if ( (minWhole == nil), {
            minWhole = 0;
        });
        if ( (leadFill == nil), {
            leadFill = " ";
        });
        dec = maxDec;
        dec.do({
            v = v * 10;
        });
        v = v.round;
        (maxDec-minDec).do({ // maybe show less decimals than maxDec?
            if ( ((v % 10) == 0), {
                v = v / 10;
                dec = dec - 1;
            });
        });
        s = v.asString;
        inspos = s.size - dec; // position where to textually insert comma
        if ( ( inspos <= 0 ), { // if s=="0", inspos may be < 0
            s = ( Toolbox.repeat("0", ( 1 - inspos ) ) ++ s ); // add leading zeros (so that "0" * 1000 = "0000")
            inspos = 1;
        });
        if ( ((dec > 1) || (minDec > 0)), { // there is at least one non-zero decimal after the comma, or at least one to show
            s = s.insert(inspos, $.);
            s = s ++ Toolbox.repeat("0", (minDec - dec)); // -inspos?
        },{ // else: all decimals are zero, and none to display: display integer format without comma
            //nop
        });
        s = Toolbox.repeat(leadFill, (minWhole - inspos)) ++ s; // add leading zeros/blanks, if requested
        ^s; // type  String
    } // end numberFormat        


    // --- countDigits(num) : int ---------------------------------------------
    //      
    *countDigits { arg num; // type int         
        var n, digits;
        digits = 0;
        n = num.round;
        while ( { n >= 1 } , {
            n = n / 10;
            digits = digits + 1;
        });
        ^digits; // type  int
    } // end countDigits        


    // --- blanks(n) : String -------------------------------------------------
    //      
    *blanks { arg n; // type int         
        ^Toolbox.repeat(" ", n); // type  String
    } // end blanks        


} // end Toolbox
